import * as pdfjs_dist from 'pdfjs-dist';
import { PDFDocumentProxy, RenderTask, PDFPageProxy } from 'pdfjs-dist';
import { ReadonlyTeleBox, NetlessApp, WindowManager } from '@netless/window-manager';

/**
 * Execute task. Optionally returns a disposer function that cleans up side effects.
 */
interface AsyncSeqFn {
    (): void | Promise<void> | (() => unknown | Promise<unknown>) | Promise<() => unknown | Promise<unknown>>;
}
interface AsyncSeqOptions {
    /**
     * Max size of the sequence.
     */
    window?: number;
    /**
     * Tail is dropped by default if the sequence is full.
     * Set this to true to drop the head instead.
     */
    dropHead?: boolean;
}
/**
 * Run async functions in sequence.
 */
declare class AsyncSeq {
    private readonly window_;
    private readonly dropHead_;
    private readonly fns_;
    private pRunning_?;
    private disposer_?;
    constructor(options?: AsyncSeqOptions);
    /**
     * Is sequence running.
     */
    get running(): boolean;
    /**
     * Size of pending tasks in the sequence.
     */
    get size(): number;
    /**
     * Is sequence full.
     */
    get full(): boolean;
    /**
     * Add task executors to the sequence.
     * @param fns Task executors. Optionally returns a disposer function that cleans up side effects.
     */
    add(...fns: AsyncSeqFn[]): Promise<void>;
    /**
     * Dispose the sequence.
     */
    dispose(): Promise<void>;
    private run_;
}

/**
 * A function that can be called to dispose resources.
 */
type Disposer<T = any> = () => T;
/**
 * An object that has a `dispose` method to dispose resources.
 */
interface IDisposable<T = any> {
    dispose(): T;
}
/**
 * A union type of {@link Disposer} and {@link IDisposable} which can be called to dispose resources.
 */
type DisposableType<T = any> = Disposer<T> | IDisposable<T>;
/**
 * A combination of {@link Disposer} and {@link IDisposable}.
 * It can be called to dispose resources or call the `dispose` method to dispose resources.
 */
interface DisposableDisposer<T = any> {
    (): T;
    dispose(): T;
}

/**
 * A Disposable Store is an {@link IDisposable} that manages {@link Disposer}s and {@link IDisposable}s.
 *
 * All {@link Disposer}s and {@link IDisposable}s in the store will be invoked(`flush`) when the store is disposed.
 *
 * A {@link DisposableStore} is also a {@link Disposer}, which means it can be the `dispose` method of an {@link IDisposable}.
 *
 * A {@link DisposableStore} is also an {@link IDisposable}, which means it can be managed by another {@link DisposableStore}.
 *
 */
interface DisposableStore extends DisposableDisposer {
    /**
     * Flush and clear all of the {@link Disposer}s and {@link IDisposable}s in the store.
     */
    (): void;
    /**
     * Get the number of {@link DisposableType}s in the store.
     */
    size(): number;
    /**
     * Add a {@link DisposableType} to the store.
     *
     * Do nothing if the {@link DisposableType} is already in the store.
     *
     * @param disposable A {@link DisposableType} .
     * @returns The same {@link DisposableType} .
     */
    add<T extends DisposableType>(disposable: T): T;
    /**
     * Add multiple {@link DisposableType}s to the store.
     *
     * Do nothing if a {@link DisposableType} is already in the store.
     *
     * @param disposables An array of {@link DisposableType}s.
     * @returns The same array of {@link DisposableType}s.
     */
    add<T extends DisposableType[]>(disposables: T): T;
    /**
     * Add each {@link DisposableType} to the store.
     *
     * Do nothing if a {@link DisposableType} is already in the store.
     *
     * @param disposables An array of {@link DisposableType}s.
     * @returns The same array of {@link DisposableType}s.
     */
    add<T extends DisposableType>(disposables: T | T[]): T | T[];
    /**
     * Invoke the executor function and add the returned {@link DisposableType} to the store.
     *
     * Do nothing if the {@link DisposableType} is already in the store.
     *
     * @param executor A function that returns a {@link DisposableType}.
     * @returns The returned {@link DisposableType}.
     */
    make<T extends DisposableType>(executor: () => T): T;
    /**
     * Invoke the executor function and add the returned {@link DisposableType} to the store.
     *
     * Do nothing if `null | undefined` is returned or the returned {@link DisposableType} is already in the store.
     *
     * @param executor A function that returns either a {@link DisposableType} or `null`.
     * @returns The returned {@link DisposableType}, or `undefined` if the executor returns `null`.
     */
    make<T extends DisposableType>(executor: () => T | null | undefined | void): T | void;
    /**
     * Invoke the executor function and add each {@link DisposableType} in the returned array to the store.
     *
     * Do nothing if a {@link DisposableType} is already in the store.
     *
     * @param executor A function that returns an array of {@link DisposableType}s.
     * @returns The returned array of {@link DisposableType}s.
     */
    make<T extends DisposableType[]>(executor: () => T): T;
    /**
     * Invoke the executor function and the returned {@link DisposableType}s to the store. Do nothing if `undefined | null` is returned.
     *
     * Do nothing if a {@link DisposableType} is already in the store.
     *
     * @param executor A function that returns either an array of {@link DisposableType}s or `undefined | null`.
     * @returns The returned array of {@link DisposableType}s, or `undefined` if the executor returns `undefined | null`.
     */
    make<T extends DisposableType[]>(executor: () => T | null | void | undefined): T | void;
    /**
     * Invoke the executor function and the returned {@link DisposableType}s to the store. Do nothing if `undefined | null` is returned.
     *
     * Do nothing if a {@link DisposableType} is already in the store.
     *
     * @param executor A function that returns either an array of {@link DisposableType}s or `undefined | null`.
     * @returns The returned array of {@link DisposableType}s, or `undefined` if the executor returns `undefined | null`.
     */
    make<T extends DisposableType>(executor: () => T | T[] | null | void | undefined): T | T[] | void;
    /**
     * Check if a {@link DisposableType} is in the store.
     *
     * @param disposable The {@link DisposableType}.
     * @returns `true` if the {@link DisposableType} is in the store, otherwise `false`.
     */
    has(disposable: DisposableType): boolean;
    /**
     * Remove the {@link DisposableType} from the store. Does not invoke the removed {@link DisposableType}.
     *
     * @param disposable The {@link DisposableType} to be flushed.
     * @returns `true` if the {@link DisposableType} is found and removed, otherwise `false`.
     */
    remove(disposable: DisposableType): boolean;
    /**
     * Invoke the {@link DisposableType} and remove it from the store at the specific key.
     *
     * @param disposable The {@link DisposableType} to be flushed. Flush all if omitted.
     */
    flush(disposable?: DisposableType): void;
    /**
     * Flush and clear all of the {@link Disposer}s and {@link IDisposable}s in the store.
     */
    dispose(this: void): void;
}

type PDFjsModule = typeof pdfjs_dist;
type GetDocument = PDFjsModule['getDocument'];
interface PDFViewerOptions {
    /** The static URL to the PDF file. */
    readonly src: Parameters<GetDocument>[0];
    /** URL to load the PDF.js lib, default is `"https://cdn.jsdelivr.net/npm/pdfjs-dist@latest/build/pdf.min.mjs"`. */
    readonly pdfjsLib?: string;
    /** URL to load the PDF.js worker, default is `"https://cdn.jsdelivr.net/npm/pdfjs-dist@latest/build/pdf.worker.min.mjs"`. */
    readonly workerSrc?: string;
    /** Default is `1.5`. */
    readonly scale?: number;
    /**
     * Whether to multiply [`devicePixelRatio`](https://developer.mozilla.org/en-US/docs/Web/API/Window/devicePixelRatio) on `scale`, default is `true`.
     * If set as a number, the final scale will be `Math.min(hidpi, devicePixelRatio) * scale`.
     */
    readonly hidpi?: boolean | number;
    /** Max rendering page size in pixels, default is no limit. If set, all pages will be shrinked to let the biggest page fit this size.  */
    readonly maxSize?: {
        readonly width: number;
        readonly height: number;
    } | null;
    /** Multiplies `scale` to render the preview pages. Default is `0.2`. */
    readonly previewScale?: number;
    /** Readonly mode (disable footer), default is `false`. */
    readonly readonly?: boolean;
}
interface PreviewLazyLoadOptions {
    container: HTMLElement;
    elements_selector: string;
    touch(element: Element | null): void;
}
declare class PreviewLazyLoad {
    readonly options: PreviewLazyLoadOptions;
    observer: IntersectionObserver;
    constructor(options: PreviewLazyLoadOptions);
    handler(entries: IntersectionObserverEntry[]): void;
    destroy(): void;
}
declare class PDFViewer implements IDisposable<void> {
    readonly options: Required<PDFViewerOptions>;
    readonly pdfjsLib: Promise<PDFjsModule>;
    readonly ready: Promise<void>;
    readonly namespace = "netless-app-pdfjs";
    readonly dispose: DisposableStore;
    readonly rendering: AsyncSeq;
    destroyed: boolean;
    getDocumentTask: ReturnType<GetDocument> | null;
    pdf: PDFDocumentProxy | null;
    lastRenderingTask: RenderTask | null;
    /** The actual `scale` shrinked by `options.maxSize`. Only valid after `ready`. */
    scale: number;
    pdfjs: PDFjsModule | null;
    dom: Element | DocumentFragment;
    contentDOM: HTMLDivElement;
    previewDOM: HTMLDivElement;
    pageDOM: HTMLDivElement;
    canvas: HTMLCanvasElement;
    whiteboardDOM: HTMLDivElement;
    footerDOM: HTMLDivElement;
    totalPageDOM: HTMLSpanElement;
    pageNumberInputDOM: HTMLInputElement;
    readonly: boolean;
    showPreview: boolean;
    previewLazyLoad: PreviewLazyLoad | null;
    touched: {
        [page: number]: true;
    };
    pageIndex: number;
    constructor(options: PDFViewerOptions);
    setDOM(dom: Element | DocumentFragment): void;
    setReadonly(readonly: boolean): void;
    setPageIndex(pageIndex: number): void;
    onNewPageIndex(index: number, _origin: "navigation" | "keydown" | "input" | "preview"): void;
    page(): Promise<PDFPageProxy> | undefined;
    size(): {
        width: number;
        height: number;
    };
    numPages(): number;
    togglePreview(show?: boolean): void;
    refresh(): void;
    onLoad(done: () => void, pdf: PDFDocumentProxy): void;
    onError(reason: unknown): void;
    onRenderStart(): void;
    onRenderEnd(): void;
    onRenderError(reason: unknown): void;
    makePreview(): Promise<void>;
    render(pageNumber: number, canvas?: HTMLCanvasElement, scale?: number): Promise<void>;
    render_(pageNumber: number, canvas?: HTMLCanvasElement, scale?: number): Promise<void>;
    private c;
    private isEditable;
}

declare const __inline: string;

type Logger = (...daga: any[]) => void;
interface PDFjsAppAttributes {
    /** The static URL to the PDF file. */
    src: string;
}
interface PDFjsAppOptions extends Pick<PDFViewerOptions, "pdfjsLib" | "workerSrc" | "scale" | "hidpi" | "maxSize" | "previewScale"> {
    /** Disables user move / scale the PDF and whiteboard. */
    disableCameraTransform?: boolean;
    /** Max scale = `maxCameraScale` * default scale. Not working when `disableCameraTransform` is true. Default: `3` */
    maxCameraScale?: number;
    /** Custom logger. Default: a logger that reports to the whiteboard server. */
    readonly log?: Logger;
}
declare class AppPDFViewer extends PDFViewer {
    log?: Logger;
    box?: ReadonlyTeleBox;
    scaleDocsToFit?: () => void;
    syncPDFView?: () => void;
    readonly jumpPage: (index: number) => void;
    constructor(options: PDFViewerOptions & {
        jumpPage: (index: number) => void;
    });
    onNewPageIndex(index: number, origin: 'navigation' | 'keydown' | 'input' | 'preview'): void;
    onRenderEnd(): void;
    onRenderError(reason: unknown): void;
}
declare const NetlessAppPDFjs: NetlessApp<PDFjsAppAttributes, {}, PDFjsAppOptions, AppPDFViewer>;
interface InstallOptions {
    /**
     * Register as another app "kind". The default kind is "PDFjs".
     */
    as?: string;
    /**
     * Options to customize the local app (not synced to others).
     */
    appOptions?: PDFjsAppOptions;
}
type RegisterFn = typeof WindowManager["register"];
/**
 * Call `register({ kind: "PDFjs", src: NetlessAppPDFjs })` to register this app.
 * Optionally accepts an options object to override the default kind.
 *
 * @example install(register, { as: "PDFViewer" })
 */
declare const install: (register: RegisterFn, options?: InstallOptions) => Promise<void>;

declare const version: string;

export { AppPDFViewer, type InstallOptions, type Logger, NetlessAppPDFjs, PDFViewer, type PDFViewerOptions, type PDFjsAppAttributes, type PDFjsAppOptions, type RegisterFn, NetlessAppPDFjs as default, install, __inline as styles, version };
